;;; extend.ss
;;; 1/28/92

;;; Copyright (C) 1987, 1992 R. Kent Dybvig
;;; Permission to copy this software, in whole or in part, to use this
;;; software for any lawful purpose, and to redistribute this software
;;; is granted subject to the restriction that all copies made of this
;;; software must include this copyright notice in full.

;;; The basic design of extend-syntax is due to Eugene Kohlbecker.  See
;;; "E. Kohlbecker: Syntactic Extensions in the Programming Language Lisp",
;;; Ph.D.  Dissertation, Indiana University, 1986."  The structure of "with"
;;; pattern/value clauses, the method for compiling extend-syntax into
;;; Scheme code, and the actual implementation are due to Kent Dybvig.

;;; Known non-ieee/r4rs features used:

;;; (when e0 e1 e2 ...)
;;; Can be defined as:
;;; (extend-syntax (when)
;;;    [(when e0 e1 e2 ...) (if e0 (begin e1 e2 ...))])

;;; (unless e0 e1 e2 ...)
;;; Can be defined as:
;;; (extend-syntax (unless)
;;;    [(unless e0 e1 e2 ...) (if (not e0) (begin e1 e2 ...))])

;;; (error who format-string format-arg ...)

;;; The symbol \\, used as an escape in syntax-match? patterns
;;; Tf your system doesn't allow this as a symbol, replace all
;;; occurrences with any other symbol besides * or ...

;;; matched [ and ] pairs used as parentheses


(let ()
   (define id
      (lambda (name access control)
         (list name access control)))
   (define id-name car)
   (define id-access cadr)
   (define id-control caddr)

   (define loop
      (lambda ()
         (vector '())))
   (define loop-ids (lambda (x) (vector-ref x 0)))
   (define loop-ids! (lambda (x v) (vector-set! x 0 v)))

   (define c...rs
      `((car caar . cdar)
        (cdr cadr . cddr)
        (caar caaar . cdaar)
        (cadr caadr . cdadr)
        (cdar cadar . cddar)
        (cddr caddr . cdddr)
        (caaar caaaar . cdaaar)
        (caadr caaadr . cdaadr)
        (cadar caadar . cdadar)
        (caddr caaddr . cdaddr)
        (cdaar cadaar . cddaar)
        (cdadr cadadr . cddadr)
        (cddar caddar . cdddar)
        (cdddr cadddr . cddddr)))

   (define add-car
      (lambda (access)
         (let ([x (and (pair? access) (assq (car access) c...rs))])
            (if x
                `(,(cadr x) ,@(cdr access))
                `(car ,access)))))

   (define add-cdr
      (lambda (access)
         (let ([x (and (pair? access) (assq (car access) c...rs))])
            (if x
                `(,(cddr x) ,@(cdr access))
                `(cdr ,access)))))

   (define duplicate-symbols
      (lambda (vars)
         (if (null? vars)
             '()
             (if (memq (car vars) (cdr vars))
                 (cons (car vars)
                       (duplicate-symbols (remove (car vars) (cdr vars))))
                 (duplicate-symbols (cdr vars))))))

   (define checkpat
      (lambda (keys pat exp)
         (let ([vars (let f ([x pat] [vars '()])
                        (cond
                           [(pair? x)
                            (if (and (pair? (cdr x))
                                     (eq? (cadr x) '...)
                                     (null? (cddr x)))
                                (f (car x) vars)
                                (f (car x) (f (cdr x) vars)))]
                           [(symbol? x)
                            (cond
                               [(memq x keys) vars]
                               [(or (eq? x 'with) (eq? x '...))
                                (error 'extend-syntax
                                       "invalid context for ~s in ~s"
                                       x exp)]
                               [else (cons x vars)])]
                           [else vars]))])
            (let ([dupls (duplicate-symbols vars)])
               (unless (null? dupls)
                  (error 'extend-syntax
                         "duplicate pattern variable name ~s in ~s"
                         (car dupls)
                         exp))))))
   (define parse
      (lambda (keys pat acc cntl ids)
         (cond
            [(symbol? pat)
             (if (memq pat keys)
                 ids
                 (cons (id pat acc cntl) ids))]
            [(pair? pat)
             (cons (id pat acc cntl)
                   (if (equal? (cdr pat) '(...))
                       (let ([x (gensym)])
                          (parse keys (car pat) x (id x acc cntl) ids))
                       (parse keys (car pat) (add-car acc) cntl
                          (parse keys (cdr pat) (add-cdr acc) cntl ids))))]
            [else ids])))

   (define pattern-variable?
      (lambda (sym ids)
         (memq sym (map id-name ids))))

   (define gen
      (lambda (keys exp ids loops qqlev)
         (cond
            [(lookup exp ids) =>
             (lambda (id)
                (add-control! (id-control id) loops)
                (list 'unquote (id-access id)))]
            [(memq exp '(quasiquote unquote unquote-splicing))
             (list 'unquote (list 'quote exp))]
            [(not (pair? exp)) exp]
            [else
             (cond
                [(and (syntax-match? '(quasiquote *) exp)
                      (not (pattern-variable? 'quasiquote ids)))
                 (list 'unquote
                       (list 'list
                             ''quasiquote
                             (make-quasi
                                (gen keys (cadr exp) ids loops
                                     (if (= qqlev 0) 0 (+ qqlev 1))))))]
                [(and (syntax-match? '(* *) exp)
                      (memq (car exp) '(unquote unquote-splicing))
                      (not (pattern-variable? (car exp) ids)))
                 (if (= qqlev 1)
                     (list (car exp) (gen-quotes keys (cadr exp) ids loops))
                     (list 'unquote
                           `(list ',(car exp)
                                    ,(make-quasi
                                        (gen keys (cadr exp) ids loops
                                             (- qqlev 1))))))]
                [(and (eq? (car exp) 'with)
                      (not (pattern-variable? 'with ids)))
                 (unless (syntax-match? '(with ((* *) ...) *) exp)
                    (error 'extend-syntax "invalid 'with' form ~s" exp))
                 (checkpat keys (map car (cadr exp)) exp)
                 (list 'unquote
                    (gen-with
                       keys
                       (map car (cadr exp))
                       (map cadr (cadr exp))
                       (caddr exp)
                       ids
                       loops))]
                [(and (pair? (cdr exp)) (eq? (cadr exp) '...))
                 (let ([x (loop)])
                    (gen-cons (list 'unquote-splicing
                                    (make-loop x (gen keys (car exp) ids
                                                      (cons x loops) qqlev)))
                              (gen keys (cddr exp) ids loops qqlev)))]
                [else
                 (gen-cons (gen keys (car exp) ids loops qqlev)
                           (gen keys (cdr exp) ids loops qqlev))])])))

   (define gen-cons
      (lambda (head tail)
         (if (null? tail)
             (if (syntax-match? '(unquote-splicing *) head)
                 (list 'unquote (cadr head))
                 (cons head tail))
             (if (syntax-match? '(unquote *) tail)
                 (list head (list 'unquote-splicing (cadr tail)))
                 (cons head tail)))))

   (define gen-with
      (lambda (keys pats exps body ids loops)
         (let ([temps (map (lambda (x) (gensym)) pats)])
            `(let (,@(map (lambda (t e) `[,t ,(gen-quotes keys e ids loops)])
                          temps
                          exps))
                ,@(let f ([ps pats] [ts temps])
                     (if (null? ps)
                         (let f ([pats pats] [temps temps] [ids ids])
                            (if (null? pats)
                                `(,(make-quasi (gen keys body ids loops 0)))
                                (f (cdr pats)
                                   (cdr temps)
                                   (parse '() (car pats) (car temps) '() ids))))
                         (let ([m (match-pattern '() (car ps))])
                            (if (eq? m '*)
                                (f (cdr ps) (cdr ts))
                                `((unless (syntax-match? ',m ,(car ts))
                                     (error
                                        ',(car keys)
                                        "~s does not fit 'with' pattern ~s"
                                        ,(car ts)
                                        ',(car ps)))
                                  ,@(f (cdr ps) (cdr ts)))))))))))

   (define gen-quotes
      (lambda (keys exp ids loops)
         (cond
            [(syntax-match? '(quote *) exp)
             (make-quasi (gen keys (cadr exp) ids loops 0))]
            [(syntax-match? '(quasiquote *) exp)
             (make-quasi (gen keys (cadr exp) ids loops 1))]
            [(pair? exp)
             (let f ([exp exp])
                (if (pair? exp)
                    (cons (gen-quotes keys (car exp) ids loops)
                          (f (cdr exp)))
                    (gen-quotes keys exp ids loops)))]
            [else exp])))

   (define lookup
      (lambda (exp ids)
         (let loop ([ls ids])
            (cond
               [(null? ls) #f]
               [(equal? (id-name (car ls)) exp) (car ls)]
               [(subexp? (id-name (car ls)) exp) #f]
               [else (loop (cdr ls))]))))

   (define subexp?
      (lambda (exp1 exp2)
         (and (symbol? exp1)
              (let f ([exp2 exp2])
                 (or (eq? exp1 exp2)
                     (and (pair? exp2)
                          (or (f (car exp2))
                              (f (cdr exp2)))))))))

   (define add-control!
      (lambda (id loops)
         (unless (null? id)
            (when (null? loops)
               (error 'extend-syntax "missing ellipsis in expansion"))
            (let ([x (loop-ids (car loops))])
               (unless (memq id x)
                  (loop-ids! (car loops) (cons id x))))
            (add-control! (id-control id) (cdr loops)))))

   (define make-loop
      (lambda (loop body)
         (let ([ids (loop-ids loop)])
            (when (null? ids)
               (error 'extend-syntax "extra ellipsis in expansion"))
            (cond
               [(equal? body (list 'unquote (id-name (car ids))))
                (id-access (car ids))]
               [(and (null? (cdr ids))
                     (syntax-match? '(unquote (* *)) body)
                     (eq? (cadadr body) (id-name (car ids))))
                `(map ,(caadr body) ,(id-access (car ids)))]
               [else
                `(map (lambda ,(map id-name ids) ,(make-quasi body))
                        ,@(map id-access ids))]))))

   (define match-pattern
      (lambda (keys pat)
         (cond
            [(symbol? pat)
             (if (memq pat keys)
                 (if (memq pat '(* \\ ...))
                     `(\\ ,pat)
                     pat)
                 '*)]
            [(pair? pat)
             (if (and (pair? (cdr pat))
                      (eq? (cadr pat) '...)
                      (null? (cddr pat)))
                 `(,(match-pattern keys (car pat)) ...)
                 (cons (match-pattern keys (car pat))
                       (match-pattern keys (cdr pat))))]
            [else pat])))

   (define make-quasi
      (lambda (exp)
         (if (and (pair? exp) (eq? (car exp) 'unquote))
             (cadr exp)
             (list 'quasiquote exp))))

   (define match-check
      (lambda (keys pat x)
         `(syntax-match? ',(match-pattern keys pat) ,x)))

   (define make-clause
      (lambda (keys cl x)
         (cond
            [(syntax-match? '(* * *) cl)
             (let ([pat (car cl)] [fender (cadr cl)] [exp (caddr cl)])
                (checkpat keys pat pat)
                (let ([ids (parse keys pat x '() '())])
                   `((and ,(match-check keys pat x)
                          ,(gen-quotes keys fender ids '()))
                     ,(make-quasi (gen keys exp ids '() 0)))))]
            [(syntax-match? '(* *) cl)
             (let ([pat (car cl)] [exp (cadr cl)])
                (checkpat keys pat pat)
                (let ([ids (parse keys pat x '() '())])
                   `(,(match-check keys pat x)
                     ,(make-quasi (gen keys exp ids '() 0)))))]
            [else
             (error 'extend-syntax "invalid clause ~s" cl)])))

   (define make-syntax
      (let ([x (string->uninterned-symbol "x")]
            [e (string->uninterned-symbol "e")])
         (lambda (keys clauses)
            (when (memq '... keys)
               (error 'extend-syntax
                       "invalid keyword ... in keyword list ~s"
                       keys))
            `(lambda (,x ,e)
                (unless (procedure? ,e)
                   (error ',(car keys) "~s is not a procedure" ,e))
                (,e (cond
                       ,@(map (lambda (cl) (make-clause keys cl x)) clauses)
                       (else (error ',(car keys) "invalid syntax ~s" ,x)))
                    ,e)))))

   (define eps-make-syntax
      (let ([x (string->uninterned-symbol "x")])
         (lambda (e keys clauses)
            (when (memq '... keys)
               (error 'extend-syntax
                       "invalid keyword ... in keyword list ~s"
                       keys))
            `(lambda (,x ,e)
                (unless (procedure? ,e)
                   (error ',(car keys) "~s is not a procedure" ,e))
                (cond
                   ,@(map (lambda (cl) (make-clause keys cl x)) clauses)
                   (else (error ',(car keys) "invalid syntax ~s" ,x)))))))

   (extend-syntax (extend-infer-syntax)
      [(extend-syntax (key1 key2 ...) clause ...)
       (andmap symbol? '(key1 key2 ...))
       (with ([proc (make-syntax '(key1 key2 ...) '(clause ...))])
          (install-syntax-expander 'key1 proc))])

   (extend-syntax (extend-syntax/code)
      [(extend-syntax/code (key1 key2 ...) clause ...)
       (andmap symbol? '(key1 key2 ...))
       (with ([proc (make-syntax '(key1 key2 ...) '(clause ...))])
          'proc)])

   (extend-syntax (extend-syntax/eps)
      [(extend-syntax/eps e (key1 key2 ...) clause ...)
       (andmap symbol? '(e key1 key2 ...))
       (with ([proc (eps-make-syntax 'e '(key1 key2 ...) '(clause ...))])
          (install-syntax-expander 'key1 proc))]))

(define syntax-match?
   (lambda (pat exp)
      (cond
         [(not (pair? pat))
          (or (eq? exp pat) (eq? pat '*))]
         [(eq? (car pat) '*)
          (if (and (pair? (cdr pat)) (eq? (cadr pat) '...))
              (let f ([lst exp])
                 (or (and (pair? lst) (f (cdr lst)))
                     (null? lst)))
              (and (pair? exp) (syntax-match? (cdr pat) (cdr exp))))]
         [(and (eq? (car pat) '\\) (pair? (cdr pat)))
          (eq? exp (cadr pat))]
         [(and (pair? (cdr pat)) (eq? (cadr pat) '...))
          (let ([pat (car pat)])
             (let f ([lst exp])
                (or (and (pair? lst)
                         (syntax-match? pat (car lst))
                         (f (cdr lst)))
                    (null? lst))))]
         [else
          (and (pair? exp)
               (syntax-match? (car pat) (car exp))
               (syntax-match? (cdr pat) (cdr exp)))])))


;;; Infer adapatation of extend-syntax package
;;; expand.ss
;;; 1/28/92

;;; Copyright (C) 1992 R. Kent Dybvig
;;; Permission to copy this software, in whole or in part, to use this
;;; software for any lawful purpose, and to redistribute this software
;;; is granted subject to the restriction that all copies made of this
;;; software must include this copyright notice in full.

;;; This file, along with the file "extend.ss", implements a complete
;;; expansion-passing-style macro expander, along with macros for all
;;; of the R4RS and IEEE Scheme syntactic forms.  See the procedures
;;; expand and partial-expand below.

;;; implementation-dependent routines

(define install-partial-syntax-expander
   (lambda (k e) (putprop k '*infer-partial-expander* e)))

(define retrieve-partial-syntax-expander
   (lambda (k) (getprop k '*infer-partial-expander*)))

(define install-syntax-expander
   (lambda (k e) (putprop k '*infer-expander* e)))

(define retrieve-syntax-expander
   (lambda (k) (getprop k '*infer-expander*)))

;;; implementation-independent code

;;; The procedure expand assumes that expanders have been defined for
;;; all forms; this is true in Chez Scheme but may not be true in some
;;; other systems.  It will expand fully into the core language used by
;;; the underlying implementation.

(define infer-expand
   (let ([initial-expander
          (lambda (x e)
             ((if (pair? x)
                  (or (and (symbol? (car x))
                           (retrieve-syntax-expander (car x)))
                      (lambda (x e) (map (lambda (x) (e x e)) x)))
                  (lambda (x e) x))
              x
              e))])
      (lambda (x) (initial-expander x initial-expander))))

(define infer-expand-once
   (let ([initial-expander
          (lambda (x e)
             ((if (pair? x)
                  (or (and (symbol? (car x))
                           (retrieve-syntax-expander (car x)))
                      (lambda (x e) (map (lambda (x) (e x e)) x)))
                  (lambda (x e) x))
              x
              e))])
      (lambda (x) (initial-expander x (lambda (x e) x)))))

;;; The procedure partial-expand looks first for partial expanders,
;;; which are defined below for all R4RS and IEEE forms.  If none
;;; exists, then it looks for a normal expander.  The partial
;;; expanders defined below leave the R4RS and IEEE forms alone
;;; but expand their contents, so the effect is to expand all
;;; and only nonstandard (user- or system-defined forms).

(define partial-expand
   (let ([initial-expander
          (lambda (x e)
             ((if (pair? x)
                  (or (and (symbol? (car x))
                           (or (retrieve-partial-syntax-expander (car x))
                               (retrieve-syntax-expander (car x))))
                      (lambda (x e) (map (lambda (x) (e x e)) x)))
                  (lambda (x e) x))
              x
              e))])
      (lambda (x) (initial-expander x initial-expander))))

(let ([install-syntax-expander install-partial-syntax-expander])

 (extend-syntax/eps e (quote)
    [(quote datum) (quote datum)])
 
 (extend-syntax/eps e (lambda)
    [(lambda formals e1 e2 ...)
     (lambda formals
        (with ([e1 (e 'e1 e)]) e1)
        (with ([e2 (e 'e2 e)]) e2)
        ...)])
 
 (extend-syntax/eps e (define)
    [(define (id . formals) e1 e2 ...)
     (define (id . formals)
        (with ([e1 (e 'e1 e)]) e1)
        (with ([e2 (e 'e2 e)]) e2)
        ...)]
    [(define id e1)
     (define id (with ([e1 (e 'e1 e)]) e1))])
 
 (extend-syntax/eps e (if)
    [(if e1 e2 e3)
     (with ([e1 (e 'e1 e)] [e2 (e 'e2 e)] [e3 (e 'e3 e)])
        (if e1 e2 e3))])
 
 (extend-syntax/eps e (set!)
    [(set! id e1)
     (with ([e1 (e 'e1 e)])
        (set! id e1))])
 
 (extend-syntax/eps e (begin)
    [(begin e1 e2 ...)
     (begin (with ([e1 (e 'e1 e)]) e1) (with ([e2 (e 'e2 e)]) e2) ...)])
 
 (extend-syntax/eps e (cond)
    [(cond (e1 e2 ...) ...)
     (cond ((with ([e1 (e 'e1 e)]) e1) (with ([e2 (e 'e2 e)]) e2) ...) ...)])
 
 (extend-syntax/eps e (and)
    [(and e1 ...)
     (and (with ([e1 (e 'e1 e)]) e1) ...)])
 
 (extend-syntax/eps e (or)
    [(or e1 ...)
     (or (with ([e1 (e 'e1 e)]) e1) ...)])
 
 (extend-syntax/eps e (case)
    [(case (keys e1 e2 ...) ...)
     (case (keys (with ([e1 (e 'e1 e)]) e1) (with ([e2 (e 'e2 e)]) e2) ...)
           ...)])
 
 (extend-syntax/eps e (let)
    [(let ((x v) ...) e1 e2 ...)
     (let ((x (with ([v (e 'v e)]) v)) ...)
        (with ([e1 (e 'e1 e)]) e1)
        (with ([e2 (e 'e2 e)]) e2)
        ...)]
    [(let f ((x v) ...) e1 e2 ...)
     (let (with ([f (e 'f e)]) f) ((x (with ([v (e 'v e)]) v)) ...)
        (with ([e1 (e 'e1 e)]) e1)
        (with ([e2 (e 'e2 e)]) e2)
        ...)])
 
 (extend-syntax/eps e (let*)
    [(let* ((x v) ...) e1 e2 ...)
     (let* ((x (with ([v (e 'v e)]) v)) ...)
        (with ([e1 (e 'e1 e)]) e1)
        (with ([e2 (e 'e2 e)]) e2)
        ...)])
 
 (extend-syntax/eps e (letrec)
    [(letrec ((x v) ...) e1 e2 ...)
     (letrec ((x (with ([v (e 'v e)]) v)) ...)
        (with ([e1 (e 'e1 e)]) e1)
        (with ([e2 (e 'e2 e)]) e2)
        ...)])
 
 (extend-syntax/eps e (do)
    [(do ((x e1 e2 ...) ...) (e3 e4 ...) e5 ...)
     (do ((x (with ([e1 (e 'e1 e)]) e1) (with ([e2 (e 'e2 e)]) e2) ...) ...)
         ((with ([e3 (e 'e3 e)]) e3) (with ([e4 (e 'e4 e)]) e4) ...)
         (with ([e5 (e 'e5 e)]) e5)
         ...)])
 
 ;; omit for IEEE Scheme
 (extend-syntax/eps e (delay)
    [(delay e1) (delay (with ([e1 (e 'e1 e)]) e1))])
 
 (let ([quasi-top
        (lambda (input e)
           (define (iskey? k x)
              (and (pair? x)
                   (eq? (car x) k)
                   (pair? (cdr x))
                   (null? (cddr x))))
           (define (quasi x n)
              (cond
                 [(pair? x)
                  (cond
                     [(iskey? 'quasiquote x)
                      (cons 'quasiquote (quasi (cdr x) (1+ n)))]
                     [(iskey? 'unquote x)
                      (if (zero? n)
                          (list 'unquote (e (cadr x) e))
                          (cons 'unquote (quasi (cdr x) (1- n))))]
                     [(and (not (zero? n)) (iskey? 'unquote-splicing x))
                      (cons 'unquote-splicing (quasi (cdr x) (1- n)))]
                     [(and (zero? n) (iskey? 'unquote-splicing (car x)))
                      (cons (list 'unquote-splicing (e (cadar x) e))
                            (quasi (cdr x) n))]
                     [else (cons (quasi (car x) n) (quasi (cdr x) n))])]
                 [(vector? x)
                  (list->vector
                     (map (lambda (x)
                             (if (iskey? 'unquote-splicing x)
                                 (list 'unquote-splicing (e (cadr x) e))
                                 (quasi x n)))
                          (vector->list x)))]
                 [(box? x) (box (quasi (unbox x) n))]
                 [(memq x '(quasiquote unquote unquote-splicing))
                  (error 'quasiquote "invalid use of ~s in `~s" x input)]
                 [else x]))
           (quasi input 0))])
    (extend-syntax/eps e (quasiquote)
       [(quasiquote e1)
        (with ([x (quasi-top 'e1 e)]) x)]))
 
 ;;; don't allow any unquote forms outside of quasiquote
 (extend-syntax/eps e (unquote))
 
 ;;; don't allow any unquote-splicing forms outside of quasiquote
 (extend-syntax/eps e (unquote-splicing))
)
